@@ -5,7 +5,7 @@
 ;; Author: Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2024
-;; Version: 0.4
+;; Version: 0.5
 ;; Package-Requires: ((emacs "28.1") (compat "30"))
 ;; URL: https://github.com/minad/dicom
 ;; Keywords: multimedia, hypermedia, files
@@ -52,6 +52,7 @@
 (require 'image)
 (require 'cus-edit)
 (require 'subr-x)
+(eval-when-compile (require 'cl-lib))
 
 ;;;; Customization
 
@@ -68,21 +69,21 @@
   "Timeout for conversion."
   :type 'natnum)
 
-(defcustom dicom-field-width 25
-  "Field width."
+(defcustom dicom-parallel (num-processors)
+  "Number of parallel conversion processes."
   :type 'natnum)
 
-(defcustom dicom-field-filter
-  '( SpecificCharacterSet
-     RecordInUseFlag
-     PrivateCreator
-     FileSetConsistencyFlag
-     FileSetID
-     IconImageSequence
-     "\\`OffsetOf"
-     "UID\\'"
-     " ")
-  "List of hidden DICOM properties."
+(defcustom dicom-attribute-width 25
+  "Attribute name width."
+  :type 'natnum)
+
+(defcustom dicom-attribute-filter
+  '( FileSetConsistencyFlag FileSetID
+     IconImageSequence PrivateCreator
+     RecordInUseFlag SpecificCharacterSet
+     "\\`OffsetOf" "UID" " ")
+  "Hidden DICOM attributes.
+The list elements are either symbols or regular expressions."
   :type '(repeat (choice string symbol)))
 
 (defcustom dicom-cache-dir (expand-file-name
@@ -93,8 +94,8 @@
   :type 'string)
 
 (defcustom dicom-play-command
-  "(mpv --loop --osd-font-size=16 --osd-margin-x=0 --osd-margin-y=0 --osd-level=3 \
---osd-status-msg='fps:${container-fps} \
+  "(mpv --loop --osd-font-size=16 --osd-margin-x=0 --osd-margin-y=0 \
+--osd-level=3 --osd-status-msg='fps:${container-fps} \
 frame:${estimated-frame-number}/${estimated-frame-count} \
 progress:${percent-pos}%%' %s) & disown"
   "Video player command line."
@@ -159,10 +160,10 @@ progress:${percent-pos}%%' %s) & disown"
 (defvar-local dicom--queue nil
   "Conversion process queue in current buffer.")
 
-(defvar-local dicom--proc nil
-  "Active conversion process in current buffer.")
+(defvar-local dicom--procs nil
+  "Active conversion processes in current buffer.")
 
-(defconst dicom--thumb-placeholder
+(defconst dicom--thumb
   '( :margin 8 :type svg :width 267 :height 200
      :data "<svg xmlns='http://www.w3.org/2000/svg' width='267' height='200'>
   <rect width='267' height='200' fill='black' stroke='gray'/>
@@ -171,19 +172,11 @@ progress:${percent-pos}%%' %s) & disown"
 </svg>")
   "Thumbnail placeholder image.")
 
-(defconst dicom--large-placeholder
-  (propertize
-   " "
-   'dicom--image t
-   'pointer 'arrow
-   'display
-   '(image :margin 8 :type svg :width 800 :height 600
-           :data "<svg xmlns='http://www.w3.org/2000/svg' width='800' height='600'>
-  <rect width='800' height='600' fill='black' stroke='gray'/>
-  <line x1='0' y1='0' x2='800' y2='600' stroke='gray'/>
-  <line x1='0' y1='600' x2='800' y2='0' stroke='gray'/>
-</svg>"))
-  "Large placeholder image.")
+(defvar dicom--display-table
+  (let ((table (make-display-table)))
+    (set-display-table-slot table 'wrap ?\s)
+    table)
+  "Display table to hide wrap characters.")
 
 ;;;; Internal functions
 
@@ -199,11 +192,9 @@ progress:${percent-pos}%%' %s) & disown"
     (ignore-errors (signal-process proc 'TERM))
     (run-at-time 1 nil (lambda () (ignore-errors (delete-process proc))))))
 
-(defun dicom--put-image (pos file)
-  "Display image FILE at POS."
-  (with-silent-modifications
-    (put-text-property pos (1+ pos) 'display
-                       `(image :margin 8 :type png :file ,file))))
+(defun dicom--image-desc (file)
+  "Image descriptor for FILE."
+  `(image :margin 8 :type png :file ,file))
 
 (defun dicom--dir-p (&optional file)
   "Non-nil if FILE is a DICOMDIR."
@@ -257,13 +248,13 @@ progress:${percent-pos}%%' %s) & disown"
                 ((not (or (equal (dom-attr dom 'loaded) "no")
                           (equal (dom-attr dom 'binary) "hidden")
                           (let (case-fold-search)
-                            (string-match-p dicom-field-filter name))))))
+                            (string-match-p dicom-attribute-filter name))))))
        (cons (intern name) (replace-regexp-in-string
                             "[ \t\n^]+" " " (dom-text dom)))))
     ('sequence
      (when-let ((name (dom-attr dom 'name))
                 ((not (let (case-fold-search)
-                        (string-match-p dicom-field-filter name))))
+                        (string-match-p dicom-attribute-filter name))))
                 (children (dicom--convert-children dom)))
        (cons (intern name) children)))))
 
@@ -274,10 +265,9 @@ progress:${percent-pos}%%' %s) & disown"
                                 "--quiet" "--charset-assume"
                                 "latin-1" "--convert-to-utf8" file))
       (error "DICOM: Reading DICOM metadata with dcm2xml failed"))
-    (let ((dicom-field-filter (string-join
-                                (mapcar (lambda (x) (format "%s" x))
-                                        dicom-field-filter)
-                                "\\|")))
+    (let ((dicom-attribute-filter (mapconcat (lambda (x) (format "%s" x))
+                                             dicom-attribute-filter
+                                             "\\|")))
       (dicom--convert (dom-child-by-tag (libxml-parse-xml-region) 'data-set)))))
 
 (defun dicom--image-buffer ()
@@ -290,36 +280,38 @@ progress:${percent-pos}%%' %s) & disown"
 (defun dicom--modify-image (fun)
   "Modify image properties by FUN."
   (with-current-buffer (dicom--image-buffer)
-   (when-let ((pos (text-property-not-all (point-min) (point-max) 'dicom--image nil))
-              (image (get-text-property pos 'display)))
-     (with-silent-modifications
-       (funcall fun image)
-       (put-text-property pos (1+ pos) 'display `(image ,@(cdr image)))))))
+    (when-let ((pos (text-property-not-all (point-min) (point-max) 'dicom--image nil))
+               (image (get-text-property pos 'display)))
+      (with-silent-modifications
+        (goto-char pos)
+        (funcall fun image)
+        (put-text-property pos (1+ pos) 'display `(image ,@(cdr image)))))))
 
 (defun dicom--run (cb &rest args)
   "Run process with ARGS asynchronously and call CB when the process finished."
   (let ((default-directory "/"))
-    (setq dicom--proc
-          (make-process
+    (push (make-process
            :name "dicom"
            :command args
            :noquery t
            :filter #'ignore
            :sentinel
            (let ((buf (current-buffer)))
-             (lambda (_proc event)
+             (lambda (proc event)
                (when (buffer-live-p buf)
                  (with-current-buffer buf
-                   (setq dicom--proc nil)
+                   (cl-callf2 delq proc dicom--procs)
                    (funcall cb (string-prefix-p "finished" event))
-                   (dicom--process)))))))
+                   (dicom--process))))))
+          dicom--procs)
     (when dicom-timeout
-      (run-at-time dicom-timeout nil #'dicom--stop dicom--proc))))
+      (run-at-time dicom-timeout nil #'dicom--stop (car dicom--procs)))))
 
 (defun dicom--enqueue (&rest job)
   "Enqueue conversion JOB."
   (push job dicom--queue)
-  (unless dicom--proc (dicom--process)))
+  (when (length< dicom--procs dicom-parallel)
+    (dicom--process)))
 
 (defun dicom--process ()
   "Process conversion queue."
@@ -355,31 +347,36 @@ progress:${percent-pos}%%' %s) & disown"
             title)
     'face (list 'dicom-title (intern (format "outline-%s" level))))))
 
+(defun dicom--image-callback (tmp dst pos)
+  "Job callback closure with TMP, DST and POS."
+  (lambda (success)
+    (if success
+        (with-silent-modifications
+          (rename-file tmp dst)
+          (put-text-property pos (1+ pos) 'display (dicom--image-desc dst)))
+      (delete-file tmp))))
+
 (defun dicom--thumb (level item)
   "Insert ITEM with thumbnail at LEVEL into buffer."
   (pcase-let* ((src (expand-file-name
                      (string-replace "\\" "/" (alist-get 'ReferencedFileID item))))
                (`(,dst . ,tmp) (dicom--cache-name src))
+               (exists (file-exists-p dst))
                (pos (point))
                (tooltip (progn
                           (dicom--item level item "")
                           (buffer-substring-no-properties pos (point)))))
     (delete-region pos (point))
     (insert (propertize
-             " " 'display `(image ,@dicom--thumb-placeholder)
+             " "
+             'display (if exists (dicom--image-desc dst) `(image ,@dicom--thumb))
              'pointer 'hand
              'keymap dicom-image-map
              'dicom--file src
              'help-echo tooltip))
-    (if (file-exists-p dst)
-        (dicom--put-image pos dst)
+    (unless exists
       (dicom--enqueue
-       (lambda (success)
-         (if success
-             (progn
-               (rename-file tmp dst)
-               (dicom--put-image pos dst))
-           (delete-file tmp)))
+       (dicom--image-callback tmp dst pos)
        "dcmj2pnm" "--write-png" "--scale-y-size" "200" src tmp))))
 
 (defun dicom--item (level item &optional indent)
@@ -409,12 +406,24 @@ progress:${percent-pos}%%' %s) & disown"
      ((not (eq k 'DirectoryRecordType))
       (let* ((k (symbol-name k))
              (s k))
-        (when (> (length s) dicom-field-width)
-          (setq s (truncate-string-to-width k dicom-field-width 0 nil "…"))
+        (when (> (length s) dicom-attribute-width)
+          (setq s (truncate-string-to-width k dicom-attribute-width 0 nil "…"))
           (put-text-property 0 (length s) 'help-echo k s))
-        (setq s (string-pad s dicom-field-width))
+        (setq s (string-pad s dicom-attribute-width))
         (insert (or indent "    ") s "  " v "\n"))))))
 
+(defun dicom--placeholder (w h)
+  "Placeholder image with W and H."
+  `(image
+    :margin 8 :type svg :width ,w :height ,h
+    :data
+    ,(format
+      "<svg xmlns='http://www.w3.org/2000/svg' width='%1$s' height='%2$s'>
+  <rect width='%1$s' height='%2$s' fill='black' stroke='gray'/>
+  <line x1='0' y1='0' x2='%1$s' y2='%2$s' stroke='gray'/>
+  <line x1='0' y1='%2$s' x2='%1$s' y2='0' stroke='gray'/>
+</svg>" w h)))
+
 (defun dicom--image ()
   "Insert large image."
   (insert (propertize "\n" 'face '(:height 0.2)))
@@ -425,18 +434,22 @@ progress:${percent-pos}%%' %s) & disown"
   (when-let ((frames (alist-get 'NumberOfFrames dicom--data)))
     (dicom--button (format "Play (%s frames)" frames) #'dicom-play))
   (insert "\n" (propertize "\n" 'face '(:height 0.2)))
-  (pcase-let ((`(,dst . ,tmp) (dicom--cache-name (concat "large" dicom--file)))
-              (pos (point)))
-    (insert dicom--large-placeholder "\n")
-    (if (file-exists-p dst)
-        (dicom--put-image pos dst)
+  (pcase-let* ((`(,dst . ,tmp) (dicom--cache-name (concat "large" dicom--file)))
+               (exists (file-exists-p dst))
+               (pos (point)))
+    (insert (propertize
+             " "
+             'dicom--image t
+             'pointer 'arrow
+             'display (if exists
+                          (dicom--image-desc dst)
+                        (dicom--placeholder
+                         (alist-get 'Columns dicom--data 800)
+                         (alist-get 'Rows dicom--data 600))))
+            "\n")
+    (unless exists
       (dicom--enqueue
-       (lambda (success)
-         (if success
-             (progn
-               (rename-file tmp dst)
-               (dicom--put-image pos dst))
-           (delete-file tmp)))
+       (dicom--image-callback tmp dst pos)
        "dcmj2pnm" "--write-png" dicom--file tmp))))
 
 (defun dicom--setup-check ()
@@ -458,15 +471,16 @@ progress:${percent-pos}%%' %s) & disown"
 (defun dicom--setup-locals (file)
   "Initialize buffer locals for FILE."
   (setq-local dicom--queue nil
-              dicom--proc nil
+              dicom--procs nil
               dicom--file file
               dicom--data (dicom--read file)
               buffer-read-only t
+              buffer-display-table dicom--display-table
               truncate-lines nil
+              left-fringe-width 0
+              right-fringe-width 0
               bookmark-make-record-function #'dicom--bookmark-record
               revert-buffer-function (lambda (&rest _) (dicom--setup file))
-              fringe-indicator-alist '((continuation . nil)
-                                       (truncation . nil))
               outline-regexp "\\*+"
               outline-minor-mode-cycle t
               outline-minor-mode-use-buttons 'in-margins
@@ -490,7 +504,7 @@ progress:${percent-pos}%%' %s) & disown"
   (condition-case err
       (progn
         (dicom--setup-check)
-        (dicom--stop dicom--proc)
+        (mapc #'dicom--stop dicom--procs)
         (dicom-mode)
         (dicom--setup-locals file)
         (dicom--setup-content)
@@ -526,35 +540,35 @@ progress:${percent-pos}%%' %s) & disown"
   "Play DICOM multi frame image."
   (interactive nil dicom-mode)
   (with-current-buffer (dicom--image-buffer)
-   (pcase-let ((`(,dst . ,tmp) (dicom--cache-name dicom--file "mp4")))
-     (cond
-      ((file-exists-p dst)
-       (message "Playing %s…" (abbreviate-file-name dicom--file))
-       (call-process-shell-command
-        (format dicom-play-command (shell-quote-argument dst))
-        nil 0))
-      (dicom--proc
-       (message "Conversion in progress…"))
-      (t
-       (unless (alist-get 'NumberOfFrames dicom--data)
-         (user-error "DICOM: No multi frame image"))
-       (let ((rate (or (alist-get 'RecommendedDisplayFrameRate dicom--data)
-                       (alist-get 'CineRate dicom--data)
-                       25))
-             dicom-timeout)
-         (message "Converting %s…" (abbreviate-file-name dicom--file))
-         (dicom--enqueue
-          (lambda (success)
-            (if success
-                (progn
-                  (rename-file tmp dst)
-                  (dicom-play))
-              (delete-file tmp)))
-          "sh" "-c"
-          (format "dcmj2pnm --all-frames --write-bmp %s | ffmpeg -framerate %s -i - %s"
-                  (shell-quote-argument dicom--file)
-                  rate
-                  (shell-quote-argument tmp)))))))))
+    (pcase-let ((`(,dst . ,tmp) (dicom--cache-name dicom--file "mp4")))
+      (cond
+       ((file-exists-p dst)
+        (message "Playing %s…" (abbreviate-file-name dicom--file))
+        (call-process-shell-command
+         (format dicom-play-command (shell-quote-argument dst))
+         nil 0))
+       (dicom--procs
+        (message "Conversion in progress…"))
+       (t
+        (unless (alist-get 'NumberOfFrames dicom--data)
+          (user-error "DICOM: No multi frame image"))
+        (message "Converting %s…" (abbreviate-file-name dicom--file))
+        (let (dicom-timeout)
+          (dicom--enqueue
+           (lambda (success)
+             (if success
+                 (progn
+                   (rename-file tmp dst)
+                   (dicom-play))
+               (delete-file tmp)))
+           "sh" "-c"
+           (format
+            "dcmj2pnm --all-frames --write-bmp %s | ffmpeg -framerate %s -i - %s"
+            (shell-quote-argument dicom--file)
+            (or (alist-get 'RecommendedDisplayFrameRate dicom--data)
+                (alist-get 'CineRate dicom--data)
+                25)
+            (shell-quote-argument tmp)))))))))
 
 ;;;###autoload
 (defun dicom-open-at-point ()
@@ -614,5 +628,21 @@ REUSE can be a buffer name to reuse."
   (add-to-list 'auto-mode-alist '("\\.\\(?:dcm\\|ima\\)\\'" . dicom-auto-mode))
   (add-to-list 'auto-mode-alist '("DICOMDIR" . dicom-auto-mode)))
 
+;;;###autoload
+(eval-after-load 'ol
+  (lambda ()
+    (defvar dicom--file)
+    (declare-function org-link-set-parameters "ol")
+    (declare-function org-link-store-props "ol")
+    (org-link-set-parameters
+     "dicom"
+     :follow (lambda (link _) (dicom-open link))
+     :store
+     (lambda ()
+       (when (derived-mode-p 'dicom-mode)
+         (org-link-store-props
+          :type "dicom"
+          :link (concat "dicom:" (abbreviate-file-name dicom--file))))))))
+
 (provide 'dicom)
 ;;; dicom.el ends here
